
from .object import make_wrapped_object, WrappedObject
from .class_ import make_wrapped_class, WrappedClassBase
from .function import make_wrapped_function
from .module import WrappedModule, WrappedSourceModule, WrappedIndirectModule
